import numpy as np
import rebound as rb
import reboundx as rbx

def set_time_step(sim,dtfactor):
    r"""
    Set the time step of a simulation to a fraction,``dtfactor``,
    of the minimum perihelion passage timescale, defined as
    
    .. math::
        \tau_{p} = P\sqrt{\frac{(1-e)^3}{1+e}}~.
    
    Generally, the time step of `WHFAST` should be set to 
    :math:`\lesssim\tau_p/20` in order to obtain reliable
    results (see :ref:`Wisdom 2015
    <https://ui.adsabs.harvard.edu/abs/2015AJ....150..127W/abstract>`)
    """
    ps=sim.particles[1:sim.N_real]
    tperi=np.min([p.P * (1-p.e)**1.5 / np.sqrt(1+p.e) for p in ps])
    dt = tperi * dtfactor
    sim.dt = dt
def set_min_distance(sim,rhillfactor):
    ps=sim.particles[1:sim.N_real]
    mstar = sim.particles[0].m
    rhill = np.min([ p.rhill for p in ps if p.m > 0])
    mindist = rhillfactor * rhill
    sim.exit_min_distance = mindist

def get_simarchive_integration_results(sa,coordinates='jacobi'):
    """
    Read a simulation archive and store orbital elements
    as arrays in a dictionary.

    Arguments
    ---------
    sa : rebound.SimulationArchive or str
     The simulation archive to read or the file name of the simulation
     archive file. Can also be a reboundx simulation archive.
    coordinates : str
        The coordinate system to use for calculating orbital elements. 
        This can be:
        | - 'jacobi' : Use Jacobi coordinates (including Jacobi masses)
        | - 'barycentric' : Use barycentric coordinates.
        | - 'heliocentric' : Use canonical heliocentric elements. 
        | The canonical cooridantes are heliocentric distance vectors.
        | The conjugate momenta are barycentric momenta.

    Returns
    -------
    sim_results : dict
        Dictionary containing time and orbital elements at each 
        snapshot of the simulation archive.
    """
    if type(sa) == str:
        sa = rb.SimulationArchive(sa)

    if type(sa) == rb.simulationarchive.SimulationArchive:
        return _get_rebound_simarchive_integration_results(sa,coordinates)
    elif type(sa) == rbx.simulationarchive.SimulationArchive:
        return _get_reboundx_simarchive_integration_results(sa,coordinates)
    raise TypeError("{} is not a rebound or reboundx simulation archive!".format(sa))

def _get_rebound_simarchive_integration_results(sa,coordinates):
    if coordinates == 'jacobi':
        get_orbits = lambda sim: sim.calculate_orbits(jacobi_masses=True)
    elif coordinates == 'heliocentric':
        get_orbits = get_canonical_heliocentric_orbits
    elif coordinates == 'barycentric':
        get_orbits = lambda sim: sim.calculate_orbits(sim.calculate_com())
    else: 
        raise ValueError("'Coordinates must be one of 'jacobi','heliocentric', or 'barycentric'")
    N = len(sa)
    sim0 = sa[0]
    Npl= sim0.N_real - 1
    shape = (Npl,N)
    sim_results = {
        'time':np.zeros(N),
        'P':np.zeros(shape),
        'e':np.zeros(shape),
        'l':np.zeros(shape),
        'inc':np.zeros(shape),
        'pomega':np.zeros(shape),
        'omega':np.zeros(shape),
        'Omega':np.zeros(shape),
        'a':np.zeros(shape),
        'Energy':np.zeros(N)
    }
    for i,sim in enumerate(sa):
        sim_results['time'][i] = sim.t
        orbits = get_orbits(sim)
        sim_results['Energy'][i] = sim.calculate_energy()
        for j,orbit in enumerate(orbits):
            sim_results['P'][j,i] = orbit.P
            sim_results['e'][j,i] = orbit.e
            sim_results['l'][j,i] = orbit.l
            sim_results['pomega'][j,i] = orbit.pomega
            sim_results['a'][j,i] = orbit.a
            sim_results['omega'][j,i] = orbit.omega
            sim_results['Omega'][j,i] = orbit.Omega
            sim_results['inc'][j,i] = orbit.inc
    return sim_results

def _get_reboundx_simarchive_integration_results(sa,coordinates):
    if coordinates == 'jacobi':
        get_orbits = lambda sim: sim.calculate_orbits(jacobi_masses=True)
    elif coordinates == 'heliocentric':
        get_orbits = get_canonical_heliocentric_orbits
    elif coordinates == 'barycentric':
        get_orbits = lambda sim: sim.calculate_orbits(sim.calculate_com())
    else: 
       raise ValueError("'Coordinates must be one of 'jacobi','heliocentric', or 'barycentric'")
    N = len(sa)
    sim0,_ = sa[0]
    Npl= sim0.N_real - 1
    shape = (Npl,N)
    sim_results = {
        'time':np.zeros(N),
        'P':np.zeros(shape),
        'e':np.zeros(shape),
        'l':np.zeros(shape),
        'inc':np.zeros(shape),
        'pomega':np.zeros(shape),
        'omega':np.zeros(shape),
        'Omega':np.zeros(shape),
        'a':np.zeros(shape),
        'Energy':np.zeros(N)
    }
    for i,sim_extra in enumerate(sa):
        sim,extra = sim_extra
        sim_results['time'][i] = sim.t
        orbits = get_orbits(sim)
        sim_results['Energy'][i] = sim.calculate_energy()
        for j,orbit in enumerate(orbits):
            sim_results['P'][j,i] = orbit.P
            sim_results['e'][j,i] = orbit.e
            sim_results['l'][j,i] = orbit.l
            sim_results['pomega'][j,i] = orbit.pomega
            sim_results['a'][j,i] = orbit.a
            sim_results['omega'][j,i] = orbit.omega
            sim_results['Omega'][j,i] = orbit.Omega
            sim_results['inc'][j,i] = orbit.inc
    return sim_results

def get_canonical_heliocentric_orbits(sim):
    """
    Compute orbital elements in canonical
    heliocentric coordinates (e.g., `Laskar & Robutel 1995`_),
    in the center-of-mass frame.
    
    .. _Laskar & Robutel 1995: https://ui.adsabs.harvard.edu/abs/1995CeMDA..62..193L/abstract

    Arguments:
    ----------
    sim : rb.Simulation
        simulation for which to compute orbits

    Returns
    -------
    list of rebound.Orbits
        Orbits of particles in canonical heliocentric
        coordinates.
    """
    return reb_calculate_orbits(sim, coordinates="canonical heliocentric")

def reb_calculate_orbits(sim, coordinates="canonical heliocentric"):
    """
    Compute orbital elements in passed canonical coordinates

    Arguments:
    ----------
    sim : rebound.Simulation
        simulation for which to compute orbits
    coordinates: str
        Specifices the canonical coordinate system. This determines the appropriate definitions of mu and M. Options:
        'canonical heliocentric' (default): canonical heliocentric coordinates in the COM frame e.g. Laskar & Robutel 1995
        'democratic heliocentric': e.g. Duncan et al. 1998

    Returns
    -------
    list of rebound.Orbits
        Orbits of particles in passed canonical coordinates 
    """
    if coordinates not in ['canonical heliocentric', 'democratic heliocentric']:
        raise AttributeError("coordinates must either be 'canonical heliocentric' (default) or 'democratic heliocentric")

    star = sim.particles[0]
    orbits = []
    # the central body in this splitting should have star.m + mu, but 
    # this is already accounted for in REBOUND's orbit calculation 
    fictitious_star = rb.Particle(m=star.m)
    for planet in sim.particles[1:sim.N_real]:
        # Heliocentric position
        r = np.array(planet.xyz) - np.array(star.xyz)
        # For canonical heliocentric coordinates the momentum
        # is the same as the inertial momentum, pi=mi*vi
        v = np.array(planet.vxyz)

        # Mapping from (coordinate,momentum) pair to
        # orbital elments requires that the 'velocity'
        # be defined as the canonical momentum divided
        # by 'mu' appearing in the definition of the
        # Delauney variables (see, e.g., pg. 34 of
        # Morbidelli 2002).
        #
        # For Laskar & Robutel (1995)'s definition
        # of canonical action-angle pairs, this is
        # the reduced mass, so v_for_orbit = mi * vi / mu_i = (m0 + mi)/m0 * vi
        # We write it this way to remain well behaved in test particle limit
        # For DHC: v_for_orbit = mi*vi/mi = vi
       
        # Need to set right central mass for REBOUND calculation
        # for CH, set m so central M = Mstar + m
        # for DHC, m=0 so central M = Mstar
        if coordinates == 'canonical heliocentric':
            m = planet.m
            v_for_orbit = (m + star.m)/star.m*v
        elif coordinates == 'democratic heliocentric':
            m = 0
            v_for_orbit = v

        fictitious_particle =  rb.Particle(
            m=m,
            x = r[0],
            y = r[1],
            z = r[2],
            vx = v_for_orbit[0],
            vy = v_for_orbit[1],
            vz = v_for_orbit[2]
        )
        orbit = fictitious_particle.calculate_orbit(primary=fictitious_star,G = sim.G)
        orbits.append(orbit)
    return orbits

def reb_add_poincare_particle(p, sim):
    elements = {element:getattr(p,element) for element in ['a','e','inc','l','pomega','Omega']}
    reb_add_from_elements(p.m, elements, sim, p.coordinates) 

def add_canonical_heliocentric_elements_particle(m,elements,sim):
    """
    Add a new particle to a rebound simulation 
    by specifying its mass and canonical heliocentric 
    orbital elements.

    Arguments
    ---------
    m : float
        Physical mass of particle to add.
    elements : dict
        Dictionary of orbital elements for particle.
        Dictionary must contain valid set
        of orbital elements accepted by REBOUND.
    sim : rebound.Simulation
        Simulation to add particle to.
    """
    reb_add_from_elements(m, elements, sim, coordinates = 'canonical heliocentric')

def reb_add_from_elements(m,elements,sim,coordinates='canonical heliocentric'):
    """
    Add a new particle to a rebound simulation 
    by specifying its mass, orbital elements, and the set of canonical
    coordinates the elements are in.

    Arguments
    ---------
    m : float
        Physical mass of particle to add.
    elements : dict
        Dictionary of orbital elements for particle.
        Dictionary must contain valid set
        of orbital elements accepted by REBOUND.
    sim : rebound.Simulation
        Simulation to add particle to.
    coordinates: str
        Specifices the canonical coordinate system. This determines the appropriate definitions of mu and M. Options:
        'canonical heliocentric' (default): canonical heliocentric coordinates in the COM frame e.g. Laskar & Robutel 1995
        'democratic heliocentric': e.g. Duncan et al. 1998
    """
    if coordinates not in ['canonical heliocentric', 'democratic heliocentric']:
        raise AttributeError("coordinates must either be 'canonical heliocentric' (default) or 'democratic heliocentric")
    
    star = sim.particles[0]

    # Make a 2body simulation with star mass m, 
    # particle mass = mu (so REBOUND assigns central mass star.m+mu)
    # Given canonical heliocentric elements,
    # this yields xtilde = xi-xstar, vtilde = (mstar+mi)/mstar * vi
    _sim = rb.Simulation()
    _sim.G = sim.G
    _star = star.copy()
    _sim.add(m=star.m)
    # REBOUND will use central mass M = star.m + m, so need to set so that correct M set
    if coordinates == 'canonical heliocentric':
        _sim.add(m=m, **elements) # so REBOUND uses central M = Mstar + m
    elif coordinates == 'democratic heliocentric':
        _sim.add(m=0, **elements) # so REBOUND uses central M = Mstar
    _p = _sim.particles[1]
    p = rb.Particle(m=m) 
    # we cache the planet positions so we can later correct to stay in COM frame
    x = star.x + _p.x
    y = star.y + _p.y
    z = star.z + _p.z
    p.x = x
    p.y = y
    p.z = z
    # Fictitious particle provides vtilde, convert back to inertial velocities
    # vtilde = momentum/canonical mass (see get_canonical_heliocentric_orbits)
    # CH: vtilde = m*v/mu, so v = f*vtilde
    # DHC: vtilde = m*v/m, so v = vtilde
    if coordinates == 'canonical heliocentric':
        f = star.m / (m + star.m)
    elif coordinates == 'democratic heliocentric':
        f = 1.
    p.vx = f * _p.vx
    p.vy = f * _p.vy
    p.vz = f * _p.vz
    # All of these need to be the final velocities in the COM frame, so the
    # only way to do that is to compensate for the shift in COM with the star
    star.vx -= m/star.m * p.vx
    star.vy -= m/star.m * p.vy
    star.vz -= m/star.m * p.vz
    sim.add(p)
    # We now move all particles by the same offset to keep the COM at 0
    # This ensures that the xi-x0 remain the same
    Mtot = np.array([p.m for p in sim.particles[:sim.N_real]]).sum()
    for p in sim.particles[:sim.N_real]:
        p.x -= m/Mtot*x
        p.y -= m/Mtot*y
        p.z -= m/Mtot*z

def _compute_transformation_angles(sim):
    Gtot_vec = sim.calculate_angular_momentum()
    Gtot_vec = np.array(Gtot_vec)
    Gtot = np.sqrt(Gtot_vec @ Gtot_vec)
    Ghat = Gtot_vec / Gtot
    Ghat_z = Ghat[-1]
    Ghat_perp = np.sqrt(1 - Ghat_z**2)
    theta1 = np.pi/2 - np.arctan2(Ghat[1],Ghat[0])
    theta2 = np.pi/2 - np.arctan2(Ghat_z,Ghat_perp)
    return theta1,theta2

def npEulerAnglesTransform(xyz,Omega,I,omega):
    x,y,z = xyz
    s1,c1 = np.sin(omega),np.cos(omega)
    x1 = c1 * x - s1 * y
    y1 = s1 * x + c1 * y
    z1 = z

    s2,c2 = np.sin(I),np.cos(I)
    x2 = x1
    y2 = c2 * y1 - s2 * z1
    z2 = s2 * y1 + c2 * z1

    s3,c3 = np.sin(Omega),np.cos(Omega)
    x3 = c3 * x2 - s3 * y2
    y3 = s3 * x2 + c3 * y2
    z3 = z2

    return np.array([x3,y3,z3])

def align_simulation(sim):
    """
    Change particle positions and velocities
    of a rebound simulations  so that the z-axis 
    corresponds with the direction of the angular 
    momentum. 

    Arguments
    ---------
    sim : rebound.Simulation
    """
    theta1,theta2 = _compute_transformation_angles(sim)
    for p in sim.particles[:sim.N_real]:
        p.x,p.y,p.z = npEulerAnglesTransform(p.xyz,0,theta2,theta1)
        p.vx,p.vy,p.vz = npEulerAnglesTransform(p.vxyz,0,theta2,theta1) 
def _get_lhat(p):
    xyz = p.xyz
    vxyz = p.vxyz
    lvec = np.cross(xyz,vxyz)
    lhat = lvec / np.linalg.norm(lvec)
    return lhat
from itertools import combinations
def calculate_mutual_inclinations(s):
    """
    Calculate the mutual inclination between pairs
    of bodies in a simulation archive

    Arguments
    ---------
    s : rebound.Simulation or rebound.SimulationArchive

    Returns
    -------
    imutuals : dictionary
        Dictionary of mutual inclinations.
        Keys are pairs of particle numbers, (i,j),
        where i and j are integers and values are 
        arrays of mutual inclination values for that 
        key pair.
    """
    if type(s)==rb.Simulation:
        return _calculate_mutual_inc(s)
    combs = combinations(range(1,sa[0].N_real),2)
    imut = {comb:np.zeros(len(sa)) for comb in combs}
    for i,sim in enumerate(sa):
        lhats = [_get_lhat(p) for p in sim.particles[1:]]
        for comb in combinations(range(1,sa[0].N_real),2):
            j,k = comb
            cosimut = lhats[j-1] @ lhats[k-1]
            imut[comb][i] = np.arccos(cosimut)

    return imut
def _calculate_mutual_inc(sim):
    combs = combinations(range(1,sim.N_real),2)
    lhats = [_get_lhat(p) for p in sim.particles[1:]]
    return {(j,k):np.arccos(lhats[j-1]@lhats[k-1]) for j,k in combs}
    



from celmech import disturbing_function
import numpy as np
from collections import OrderedDict
from sympy import S
from itertools import combinations
import rebound

def masses_to_jacobi(masses):
    N = len(masses)
    mjac, Mjac = np.zeros(N), np.zeros(N)
    interior_mass = masses[0]
    for i in range(1,N):
        mjac[i] = masses[i]*interior_mass/(masses[i]+interior_mass) # reduced mass with interior mass
        Mjac[i] = masses[0]*(masses[i]+interior_mass)/interior_mass # mjac[i]*Mjac[i] always = m[i]*m[0]
        interior_mass += masses[i]
    mjac[0] = interior_mass
    return list(mjac), list(Mjac)

def masses_from_jacobi(mjac, Mjac):
    plus = Mjac[1] # m0+m1
    product = mjac[1]*Mjac[1] # m0m1
    minus = np.sqrt(plus**2 - 4*product) # m0-m1
    m0 = (plus + minus)/2.
    masses = [m0] + [Mjac[i]*mjac[i]/m0 for i in range(1,len(mjac))] # mjac[i]*Mjac[i] = m0mi
    return masses

def masses_to_heliocentric(masses):
    r"""
    Convert list of masses to mass parameters 
    used in the definition of canonical 
    heliocentric Delauney actions.

    Arguments
    ---------
    masses : ndarray
        Masses of particles in the system with stellar mass
        listed first.

    Returns
    -------
    mu : list
        List of reduced masses, 

        .. math::

            \mu_i = \frac{M_* m_i}{M_* + m_i},

        for :math:`i\ge 1`. 
        mu[0] is the stellar mass.

    M : list
        List of
        
        .. math::

            M_i = M_* + m_i

        values for :math:`i\ge 1`. 
        M[0] is set to 0.
    """
    N=len(masses)
    masses_arr = np.array(masses)
    mstar = masses_arr[0]
    mplanet =masses_arr[1:]
    M = mplanet + mstar
    mu = mplanet * mstar / (mstar + mplanet)
    return [mstar] + mu.tolist(), [0] + M.tolist() 

def masses_from_heliocentric(mu,M):
    r"""
    Convert list of mass parameters 
    used in the definition of canonical 
    heliocentric Delauney actions to 
    star and planet masses. 

    Arguments
    ---------
    mu : array-like
        List of reduced masses, 

        .. math::

            \mu_i = \frac{M_* m_i}{M_* + m_i},

        for :math:`i\ge 1`. 

    M : array-like
        List of
        
        .. math::

            M_i = M_* + m_i

        values for :math:`i\ge 1`. 

    Arguments
    ---------
    masses : list 
        Masses of particles in the system with stellar mass
        listed first.

    """
    mu_arr = np.array(mu)[1:]
    M_arr = np.array(M)[1:]
    X = np.sqrt(M_arr**2 - 4 * M_arr * mu_arr)
    mstar_arr = 0.5 * (M_arr + X)
    m_arr = 0.5 * (M_arr - X)
    assert np.alltrue(np.isclose(mstar_arr,mstar_arr[0],rtol=1e-10))
    mstar = np.mean(mstar_arr)
    return [mstar] + m_arr.tolist()

def ActionAngleToXY(Action,angle):
        return np.sqrt(2*Action)*np.cos(angle),np.sqrt(2*Action)*np.sin(angle)

def XYToActionAngle(X,Y):
        return 0.5 * (X*X+Y*Y), np.arctan2(Y,X)

def pol_to_cart(R, phi):
    X = R*np.cos(phi)
    Y = R*np.sin(phi)
    return X, Y

def cart_to_pol(X, Y):
    R = np.sqrt(X**2 + Y**2)
    phi = np.arctan2(Y,X)
    return R, phi




import numpy as np
from sympy import symbols, series
from scipy.special import k0,k1,p_roots
import warnings
from . import clibcelmech
from .nbody_simulation_utilities import get_canonical_heliocentric_orbits
from ctypes import POINTER,c_int,c_double,c_long

_machine_eps = np.finfo(np.float64).eps

def get_symbol(latex, subscript=None, **kwargs): # i=None, kwargs
    """
    Get a sympy sympy based on an input LaTeX string.
    Valid keyword arguments for the function ``sympy.symbols``
    can also be passed. 

    Arguments
    ---------
    latex : string
        LaTeX expression to render as a sympy symbol
    subscript : string or int, optional
        A subscript for the sympy symbol

    Returns
    -------
    sympy symbol
    """
    if subscript:
        return symbols(r"{0}_{{{1}}}".format(latex, subscript), **kwargs)
    else:
        return symbols(r"{0}".format(latex), **kwargs)

def get_symbol0(latex, subscript=None, **kwargs): # i=None, kwargs
    """
    Same as :func:`get_symbol`, but appends a "0" to the subscript.
    """
    if subscript:
        return symbols(r"{0}_{{{1}\,0}}".format(latex, subscript), **kwargs)
    else:
        return symbols(r"{0}_0".format(latex), **kwargs)

def sk(k,y,tol=1.49e-08,rtol=1.49e-08,maxiter=50,miniter=1):
    """
    Approximate disturibing function coefficient described in
    `Hadden & Lithwick (2018)`_

    .. _Hadden & Lithwick (2018): https://ui.adsabs.harvard.edu/abs/2018AJ....156...95H/abstract

    Quadrature routine based on scipy.quadrature.

    Arguments
    ---------
    k : int
        Order of resonance
    y : float
        e / ecross; must be y<1
    tol, rtol: float, optional
        Control absolute and relative tolerance of integration.
        Iteration stops when error between last two iterates 
        is less than `tol` OR the relative change is less than 
        `rtol`.
    maxiter : int, optional
        Maximum order of Gaussian quadrature.
    miniter : int, optional
        Minimum order of Gaussian quadrature

    Returns
    -------
    val : float
        Gaussian quadrature approximation of s_k(y)
    """
    if y>1:
        raise ValueError("sk(k,y) called with y={:f}."
        "Value of y must be less than 1.")
    maxiter=max(miniter+1,maxiter)
    val = np.inf
    err = np.inf
    for n in xrange(miniter,maxiter+1):
        newval = _sk_integral_fixed_quad(k,y,n)
        err = abs(newval-val)
        val = newval
        if err<tol or err< rtol*abs(val):
            break
    else:
        warnings.warn("maxiter (%d) exceeded. Latest difference = %e" % (maxiter, err))
    return val

def _sk_integral_fixed_quad(k,y,Nquad):

    # Get numerical quadrature nodes and weight
    nodes,weights = p_roots(Nquad)
    
    # Rescale for integration interval from [-1,1] to [-pi,pi]
    nodes = nodes * np.pi
    weights = weights * 0.5
    arg1 = 2 * k * (1 + y * np.cos(nodes)) / 3
    arg2 = k * nodes + 4 * k * y * np.sin(nodes) / 3
    integrand = k0(arg1) * np.cos(arg2)
    return  (2/np.pi) * integrand @ weights

def Dsk(k,y,tol=1.49e-08,rtol=1.49e-08,maxiter=50,miniter=1):
    """
    Derivative of disturibing function coefficient s_k
    with respect to argument y. Coefficients are described 
    in `Hadden & Lithwick (2018)`_

    .. _Hadden & Lithwick (2018): https://ui.adsabs.harvard.edu/abs/2018AJ....156...95H/abstract
    
    Quadrature routine based on scipy.quadrature.

    Arguments
    ---------
    k : int
        Order of resonance
    y : float
        e / ecross; must be y<1
    tol, rtol: float, optional
        Control absolute and relative tolerance of integration.
        Iteration stops when error between last two iterates 
        is less than `tol` OR the relative change is less than 
        `rtol`.
    maxiter : int, optional
        Maximum order of Gaussian quadrature.
    miniter : int, optional
        Minimum order of Gaussian quadrature

    Returns
    -------
    val : float
        Gaussian quadrature approximation of s_k(y)
    """
    if y>1:
        raise ValueError("sk(k,y) called with y={:f}."
        "Value of y must be less than 1.")
    maxiter=max(miniter+1,maxiter)
    val = np.inf
    err = np.inf
    for n in xrange(miniter,maxiter+1):
        newval = _Dsk_integral_fixed_quad(k,y,n)
        err = abs(newval-val)
        val = newval
        if err<tol or err< rtol*abs(val):
            break
    else:
        warnings.warn("maxiter (%d) exceeded. Latest difference = %e" % (maxiter, err))
    return val

def _Dsk_integral_fixed_quad(k,y,Nquad):

    # Get numerical quadrature nodes and weight
    nodes,weights = p_roots(Nquad)
    
    # Rescale for integration interval from [-1,1] to [-pi,pi]
    nodes = nodes * np.pi
    weights = weights * 0.5
    arg1 = 2 * k * (1 + y * np.cos(nodes)) / 3
    arg2 = k * nodes + 4 * k * y * np.sin(nodes) / 3
    integrand = -2 * k * k1(arg1) * np.cos(nodes) * np.cos(arg2) / 3 - 4 * k * k0(arg1) * np.sin(nodes) * np.sin(arg2) / 3
    return (2/np.pi) * integrand @ weights

def getOmegaMatrix(n):
    """
    Get the 2n x 2n skew-symmetric block matrix:
          [0 , I_n]
          [-I_n, 0 ]
    that appears in Hamilton's equations.

    Arguments
    ---------
    n : int
        Determines matrix dimension

    Returns
    -------
    numpy.array
    """
    zeros = np.zeros((n,n),dtype=int)
    I = np.eye(n,dtype=int)
    return np.vstack(
        (
         np.concatenate([zeros,I]).T,
         np.concatenate([-I,zeros]).T
        )
    )
######################################################
################ AMD Calculation #####################
######################################################
from scipy.optimize import brenth
def _F(e,alpha,gamma):
    """Equation 35 of Laskar & Petit (2017)"""
    denom = np.sqrt(alpha*(1-e*e)+gamma*gamma*e*e)
    return alpha*e -1 + alpha + gamma*e / denom

def _F_for_res_overlap(e,alpha,gamma,mutot):
    """Equation 35 of Laskar & Petit (2017)"""
    fByg = alpha**(0.825)
    ecross = 1/alpha - 1
    daBya= 1 - alpha
    ecrit = ecross * np.exp(-2.2 * mutot**(1/3) * daBya**(-4/3)) 
    denom = np.sqrt( (1 - e*e) * fByg**2 + e*e*alpha*gamma*gamma )
    e1 = gamma * np.sqrt(alpha) * e / denom 

    return fByg * e + e1 - fByg  * ecrit 

def critical_relative_AMD(alpha,gamma):
    r"""
    The critical value of 'relative AMD', :math:`{\cal C} = C/\Lambda_\mathrm{out}`,
    of a planet pair above which intersecting orbits are allowed.

    See Equation 29 of 
    `Laskar & Petit (2017) <https://ui.adsabs.harvard.edu/abs/2017A%26A...605A..72L/abstract>`_

    Arguments
    ---------
    alpha : float
        The semi-major axis ratio, :math:`\alpha=a_\mathrm{in}/a_\mathrm{out}` of the planet pair.
    gamma : float
        The mass ratio of the planet pair, :math:`\gamma = m_\mathrm{in}/m_\mathrm{out}`.

    Returns
    -------
    Ccrit : float
        The value of the the critical AMD
        (:math:`C_c(\alpha,\gamma)` in the notation of 
        `Laskar & Petit (2017) 
        <https://ui.adsabs.harvard.edu/abs/2017A%26A...605A..72L/abstract>`_
    """
    e0 = np.min((1,1/alpha-1))
    ec = brenth(_F,0,e0,args=(alpha,gamma))
    e1c = np.sin(np.arctan(gamma*ec / np.sqrt(alpha*(1-ec*ec))))
    curlyC = gamma*np.sqrt(alpha) * (1-np.sqrt(1-ec*ec)) + (1 - np.sqrt(1-e1c*e1c))
    return curlyC

def critical_relative_AMD_resonance_overlap(alpha,gamma,mutot):
    r"""
    The critical value of 'relative AMD', :math:`{\cal C} = C/\Lambda_\mathrm{out}`,
    of a planet pair above which resonance overlap can occur based
    on the resonance overlap criterion of Hadden & Lithwick (2018)

    Arguments
    ---------
    alpha : float
        The semi-major axis ratio, :math:`\alpha=a_\mathrm{in}/a_\mathrm{out}` of the planet pair.
    gamma : float
        The mass ratio of the planet pair, :math:`\gamma = m_\mathrm{in}/m_\mathrm{out}`.
    mutot : float
        The total mass of the planet pair relative to the star, i.e., 
        :math:`(\mu_\mathrm{in} + \mu_\mathrm{out}) / M_*`

    Returns
    -------
    Ccrit : float
        The value of the the critical AMD
        (:math:`C_c(\alpha,\gamma)` in the notation of 
        `Laskar & Petit (2017) 
        <https://ui.adsabs.harvard.edu/abs/2017A%26A...605A..72L/abstract>`_
    """
    e0 = np.min((1,1/alpha-1))
    ec = brenth(_F_for_res_overlap,0,e0,args=(alpha,gamma,mutot))
    fByg = alpha**(0.825)
    denom = np.sqrt( (1 - ec*ec) * fByg**2 + ec*ec*alpha*gamma*gamma )
    e1c = gamma * np.sqrt(alpha) * ec / denom 
    curlyC = gamma*np.sqrt(alpha) * (1-np.sqrt(1-ec*ec)) + (1 - np.sqrt(1-e1c*e1c))
    return curlyC

def compute_AMD(sim):
    """
    Compute total AMD of a planetary system.
    
    The angular momentum deficit (AMD) of a 
    planetary system is the difference between
    the angular momentum of a hypothetical system
    with the same masses and semi-major axes but with
    circular, coplanar orbits and the actual 
    angular momentum of a planetary system.
    It is a conserved quantity of the purely
    secular dynamics of a system.

    Arguments
    ---------
    sim : :class:`rebound.Simulation`
        A REBOUND simulation of a planetary system.

    Returns
    -------
    AMD : float
        The value of the systems angular momentum 
        deficit.
    """

    pstar = sim.particles[0]
    Mstar = pstar.m
    Ltot = pstar.m * np.cross(pstar.xyz,pstar.vxyz)
    ps = sim.particles[1:]
    Lmbda=np.zeros(len(ps))
    G = np.zeros(len(ps))
    Lhat = np.zeros((len(ps),3))
    ch_orbits = get_canonical_heliocentric_orbits(sim)
    for k,p in enumerate(sim.particles[1:]):
        orb = ch_orbits[k]
        GMi = sim.G * (p.m + Mstar)
        mu = p.m*Mstar/(p.m + Mstar)
        Lmbda[k] = mu * np.sqrt(GMi * orb.a)
        G[k] = Lmbda[k] * np.sqrt(1-orb.e*orb.e)
        hvec = np.cross(p.xyz,p.vxyz)
        Lhat[k] = hvec / np.linalg.norm(hvec)
        Ltot = Ltot + p.m * hvec
    cosi = np.array([Lh.dot(Ltot) for Lh in Lhat]) / np.linalg.norm(Ltot)
    return np.sum(Lmbda) - np.sum(G * cosi)

def AMD_stable_Q(sim):
    """
    Test the AMD-stability of a planetary system.
    Returns :code:`True` if a planetary system is AMD-stable
    and :code:`False` if not. 

    Arguments
    ---------
    sim : :class:`rebound.Simulation`
     Simulation object to copmute stability criterion for.
     
    Returns
    -------
    bool : 
     :code:`True` if the sytem is AMD-stable, otherwise :code:`False`.
    """
    AMD = compute_AMD(sim)
    pstar = sim.particles[0]
    ps = sim.particles[1:]
    for i in range(len(ps)-1):
        pIn = ps[i]
        pOut = ps[i+1]
        orbIn = pIn.calculate_orbit(pstar)
        orbOut = pOut.calculate_orbit(pstar)
        alpha = orbIn.a / orbOut.a
        gamma = pIn.m / pOut.m
        LmbdaOut = pOut.m * np.sqrt(orbOut.a)
        Ccrit = critical_relative_AMD(alpha,gamma)
        C = AMD / LmbdaOut
        if C>Ccrit:
            return False
    return True

def AMD_stability_coefficients(sim,overlap=False):
    r"""
    Compute AMD stability coefficients 
    of the successive adjacent planet pairs 
    of a planetary system.

    A planet pair's AMD stability coefficicent 
    is defined as the total planetary system's
    AMD divided by the critical AMD required
    for the pair's orbits to cross.
    (Equation 58 of `Laskar & Petit (2017) 
    <https://ui.adsabs.harvard.edu/abs/2017A%26A...605A..72L/abstract>`_)

    
    Arguments
    ---------
    sim : rebound.Simulation
      Simulation object to copmute AMD coefficients for.
    overlap : bool, optional
      If True, planet pairs' critical AMD values are computed
      as the critical AMD value for resonance overlap. By default 
      the critical values are computed as the value required
      for orbit crossing. 
     
    Returns
    -------
    ndarray : 
      Values of :math:`\beta = \frac{C}{\Lambda'C_c}`
      for planet pairs.
    """
    AMD = compute_AMD(sim)
    pstar = sim.particles[0]
    ps = sim.particles[1:]
    coeffs = np.zeros(len(ps)-1)
    for i in range(len(ps)-1):
        pIn = ps[i]
        pOut = ps[i+1]
        orbIn = pIn.calculate_orbit(pstar)
        orbOut = pOut.calculate_orbit(pstar)
        alpha = orbIn.a / orbOut.a
        gamma = pIn.m / pOut.m
        LmbdaOut = pOut.m * np.sqrt(orbOut.a)
        if overlap:
            mutot = (pIn.m + pOut.m) / pstar.m
            Ccrit = critical_relative_AMD_resonance_overlap(alpha,gamma,mutot)
        else:
            Ccrit = critical_relative_AMD(alpha,gamma)
        C = AMD / LmbdaOut
        coeffs[i] = C / Ccrit
    return coeffs

######################################################
######################## FMFT ########################
######################################################
p2d = np.ctypeslib.ndpointer(dtype = np.float,ndim = 2,flags = 'C')
_fmft = clibcelmech.fmft_wrapper
_fmft.argtypes =[p2d, c_int, c_double, c_double, c_int, p2d, c_long]
_fmft.restype = c_int
def _check_errors(ret, func, args):
    if ret<=0:
        raise RuntimeError("FMFT returned error code %d for the given arguments"%ret)
    return ret
_fmft.errcheck = _check_errors
def _nearest_pow2(x):
	return int(2**np.floor(np.log2(x)))
def frequency_modified_fourier_transform(time, z, Nfreq, method_flag = 3, min_freq = None, max_freq = None):
    """
    Apply the frequency-modified Fourier transfrorm algorithm of `Šidlichovský & Nesvorný (1996)`_
    to a time series to determine the series' principle Fourier modes. This function simply
    proivdes a wrapper to to C implementation written by D. Nesvorný available
    at `www-n.oca.eu/nesvorny/programs.html`_.


    .. _Šidlichovský & Nesvorný (1996): https://ui.adsabs.harvard.edu/abs/1996CeMDA..65..137S/abstract>
    .. _www-n.oca.eu/nesvorny/programs.html: https://www-n.oca.eu/nesvorny/programs.html

    Arguments
    ---------
    time : ndarray, shape (N,)
        Times of input data values.
    z : complex ndarray, shape (N,) 
      Input data time series in the form.
    Nfreq : int
        Number of Fourier modes to determine.
    method_flag : int
        The FMFT algorithm 
		Basic Fourier Transform algorithm           if   flag = 0;   not implemented   
		Modified Fourier Transform                  if   flag = 1;
		Frequency Modified Fourier Transform        if   flag = 2;
		FMFT with additional non-linear correction  if   flag = 3
    """
    output_arr = np.empty((Nfreq,3),order='C',dtype=np.float64)
    input_arr = np.array(np.vstack((time,np.real(z),np.imag(z))).T,order='C',dtype=np.float64)
    Ndata = _nearest_pow2(len(input_arr))
    dt = time[1]-time[0]
    _Nyq =  np.pi / dt
    if not min_freq:
        min_freq = -1 * _Nyq
    if not max_freq:
        max_freq = _Nyq
    _fmft( output_arr,
            Nfreq,
            min_freq,
            max_freq,
            c_int(method_flag),
            input_arr,
            c_long(Ndata)
    )
    return {x[0]:x[1]*np.exp(1j*x[2]) for x in output_arr}

def holman_weigert_stability_boundary(mu,e,Ptype=True):
    r"""
    Compute the critical semi-major axis represnting an approximate
    stability boundary for circumbinary planets in P- or S-type orbits.
    Formulas for critical semi-major axes are taken from `Holman & Wiegert (1999)`_

    .. _Holman & Wiegert (1999): https://ui.adsabs.harvard.edu/abs/1999AJ....117..621H/abstract

    Arguments
    ---------
    mu : float
      The mass-ratio of the binary,

      .. math::
        \mu = \frac{m_B}{m_A+m_B}

      where math:`m_A` and math:`m_B` are the component masses of the binary.
    e : float
      The eccentricity of the binary.
    Ptype : bool, optional
      If ``True`` (default) orbit is assumed to be a P-type circumbinary orbit.
      If ``False``, a S-type circum-primary/secondary orbit is considered.

    Returns
    -------
    aC : float
      The critical semi-major axis marking the stability boundary
    """
    if Ptype:
        if mu<0.1 or mu>0.5:
            warnings.warn("Input 'mu'={:.2g} is outside range [0.1,0.5] for which the stability boundary has been computed".format(mu))
        if e<0.0 or e>0.7:
            warnings.warn("Input 'e'={:.2g} is outside range [0.0,0.7] for which the stability boundary has been computed".format(e))
        aC =  1.6
        aC += 5.1 * e
        aC += -2.22 * e * e
        aC += 4.12 * mu
        aC += -4.27 * mu * e
        aC += -5.09 * mu * mu
        aC += 4.61 * mu * mu * e * e
    else:
        if mu<0.1 or mu>0.9:
            warnings.warn("Input 'mu'={:.2g} is outside range [0.1,0.5] for which the stability boundary has been computed".format(mu))
        if e<0.0 or e>0.8:
            warnings.warn("Input 'e'={:.2g} is outside range [0.0,0.7] for which the stability boundary has been computed".format(e))
        aC =  0.464
        aC += -0.38 * mu
        aC += -0.631 * e
        aC += 0.586 * mu * e
        aC += 0.150 * e * e
        aC += -0.198 * mu * e * e
    return aC
#######################
from sympy import diff, Matrix
def poisson_bracket(f,g,re_varslist,complex_varslist):
    r"""
    Calculate the Poisson bracket

    .. math::
        [f,g] = \sum_{i=1}^N
        \frac{\partial f}{\partial q_i}
        \frac{\partial g}{\partial p_i}
        -
        \frac{\partial f}{\partial p_i}
        \frac{\partial g}{\partial q_i}
        -
        i \sum_{j=1}^{M}
        \frac{\partial f}{\partial z_j}
        \frac{\partial g}{\partial \bar{z}_j}
        -
        \frac{\partial f}{\partial \bar{z}_j}
        \frac{\partial g}{\partial {z}_i}

    where :code:`re_varslist` is :math:`=(q_1,...,q_N,p_1,...,p_N)`
    and :code:`complex_varslist` is :math:`=(x_1,...,x_M,\bar{x}_1,...,\bar{x}_M)`.

    Arguments
    ---------
    f : sympy expression
        Function appearing in Poisson bracket.
    g : sympy expression
        Other function appearing in Poisson bracket.
    re_varslist : list of sympy symbols
        List of real canonical variables in the form 
        :math:`(q_1,...,q_N,p_1,...,p_N)`
    complex_varslist : list of sympy symbols
        List of complex canonical variables in the form 
        :math:`(x_1,...,x_M,\bar{x}_1,...,\bar{x}_M)`

    Returns
    -------
    sympy expression
    """
    br = 0
    if len(complex_varslist)>0:
        Omega_c =Matrix(-1j * getOmegaMatrix(len(complex_varslist)//2))
        gradf_c = Matrix([diff(f,v) for v in complex_varslist])
        gradg_c = Matrix([diff(g,v) for v in complex_varslist])
        br +=  gradf_c.dot(Omega_c * gradg_c)
    if len(re_varslist)>0:
        Omega_re=Matrix(getOmegaMatrix(len(re_varslist)//2))
        gradf_re = Matrix([diff(f,v) for v in re_varslist])
        gradg_re = Matrix([diff(g,v) for v in re_varslist])
        br+= gradf_re.dot(Omega_re * gradg_re)
    return br

def truncated_expansion(exprn,order_rules,max_order):
    r"""
    Expand a sympy expression up to a maximum order in a
    small book-keeping parameter after assigning variables 
    appearing in the expression a given order using the 
    `order_rules` argument.

    Arguments
    ---------
    exprn : sympy expression
        The original expression from which to calculate
        expansion.
    order_rules : dict
        A dictionary specifying what order various variables
        should be assumed to have in the book-keeping parameter.
        Each key-value pair ``{n:[x_1,x_2,..,x_m]}`` in ``order_rules``
        specifies that a set of variables 

        .. math::
            (x_1,...,x_m) \sim \mathcal{O}(\epsilon^n)
        
        where :math:`\epsilon` is the book-keeping parameter.
    max_order : int
        The order at which the resulting series expansion in 
        the book-keeping parameter :math:`\epsilon` should
        be truncated.

    Returns
    -------
    sympy expression
    """
    eps = symbols("epsilon")
    assert eps not in exprn.free_symbols, "Epsilon appears as a free symbols in 'exprn'."
    rule = dict()
    for n,variables in order_rules.items():
        rule.update({v:eps**n * v for v in variables})
    sexprn = series(exprn.subs(rule),eps,0,max_order+1)
    result = sexprn.removeO().subs({eps:1})
    return result



try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension
from codecs import open
import os
import sys

import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

# Try to get git hash
try:
    import subprocess
    ghash = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii")
    ghash_arg = "-DCELMECHGITHASH="+ghash.strip()
except:
    ghash_arg = "-DCELMECHGITHASH=0b07b5a7cab5243fc616c0d2dcbdb20ce47397e0" #GITHASHAUTOUPDATE

extra_link_args=[]
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-shared')
    extra_link_args=['-Wl,-install_name,@rpath/libcelmech'+suffix]
    
libcelmechmodule = Extension('libcelmech',
                    sources = [ 
                        'src/disturbing_function.c',
                        'src/poisson_series.c',
                        'src/fmft.c',
                        'src/fmftPy.c',
                        'src/nrutil.c'
                    ],
                    include_dirs = ['src'],
                    define_macros=[ ('LIBCELMECH', None) ],
                    # Removed '-march=native' for now.
                    extra_compile_args=['-fstrict-aliasing', '-O3','-std=c99','-Wno-unknown-pragmas', ghash_arg, '-DLIBCELMECH', '-D_GNU_SOURCE', '-fPIC'],
                    extra_link_args=extra_link_args,
                    )

if not os.getenv('READTHEDOCS'):
    packages = ['theano', 'sympy>=1.1.1', 'numpy', 'scipy>=1.2.0', 'reboundx>=3.1.0', 'rebound>=3.5.11', 'mpmath>=1.0.0']
    try:
        install_requires += packages
    except:
        install_requires = packages

setup(name='celmech',
    version='1.0.3',
    description='Open source tools for celestial mechanics',
    url='http://github.com/shadden/celmech',
    author='Dan Tamayo, Sam Hadden',
    author_email='tamayo.daniel@gmail.com, shadden1107@gmail.com',
    license='GPL',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='astronomy astrophysics celestial-mechanics orbits orbital-mechanics',
    packages=['celmech'],
    install_requires=['theano', 'mpmath>=1.0.0', 'sympy>=1.1.1', 'rebound>=3.5.11', 'reboundx>=3.1.0', 'numpy', 'scipy>=1.2.0'],
    test_suite="celmech.test",
    ext_modules = [libcelmechmodule],
    zip_safe=False)
